/*
    Copyright 2006-2016 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#define _LARGE_FILES        // large file support is usually automatic now
#define __USE_LARGEFILE64   // but I prefer to declare all of them
#define __USE_FILE_OFFSET64
#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS   64

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

#ifdef WIN32
    #include <direct.h>
    #include <windows.h>

    #define PATHSLASH   '\\'
    #define MKDIR(X)    mkdir(X)
#else
    #include <unistd.h>
    #include <sys/stat.h>
    #include <sys/types.h>
    #include <dirent.h>

    #define stricmp     strcasecmp
    #define stristr     strcasestr
    #define PATHSLASH   '/'
    #define MKDIR(X)    mkdir(X, 0755)
#endif

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;
typedef uint64_t    u64;

#define PRINTF64(x) (u32)(((x) >> 32) & 0xffffffff), (u32)((x) & 0xffffffff)    // I64x, llx? blah
#if defined(_LARGE_FILES)
    #if defined(__APPLE__)
        #define fseek   fseeko
        #define ftell   ftello
    #elif defined(__FreeBSD__)
    #elif !defined(NOLFS)       // use -DNOLFS if this tool can't be compiled on your OS!
        #define off_t   off64_t
        #define fopen   fopen64
        #define fseek   fseeko64
        #define ftell   ftello64
    #endif
#endif



#define VER         "0.2.1"
#define BUFFSZ      8192
int SEEKSZ =        2048;
#define FNAMEMAX    512
#define OFFSZNAME   "\n"                                \
                    " offset    length    filename\n"   \
                    " ============================\n"



void hedwad_extract(u8 *fname, u8 *folder);
void hedwad_build(u8 *fname, u8 *folder);
u32 get32(FILE *fd);
void put32(FILE *fd, u32 num);
void getfn(FILE *fd, u8 *dst, u32 dstlen);
void putfn(FILE *fd, u8 *name);
void copy_file(FILE *fd, FILE *fdo, u8 *fname, u32 flen);
int build(u8 *fname, u32 length);
u8 *create_dir(u8 *name);
int recursive_dir(u8 *filedir);
void std_err(void);



FILE    *fd_hed,
        *fd_wad;
u32     filez;
int     list_only   = 0;



int main(int argc, char *argv[]) {
    u8      *p,
            *cmd,
            *fname,
            *folder;

    setbuf(stdout, NULL);
    setbuf(stderr, NULL);

    fputs("\n"
        "THPS HED/WAD files extractor/builder "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stderr);

    if(argc < 4) {
        printf("\n"
            "Usage: %s <e/l/b> <file.HED/WAD> <folder> [SEEKSZ]\n"
            "\n"
            "use e for extracting the files in a wad archive into folder\n"
            "use l for listing the files in a wad archive into folder\n"
            "use b for building a wad file using the files contained in folder\n"
            "\n", argv[0]);
        exit(1);
    }

    cmd    = argv[1];
    fname  = argv[2];
    folder = argv[3];
    if(argc > 4) SEEKSZ = atoi(argv[4]);

    p = malloc(strlen(fname) + 5);
    if(!p) std_err();
    strcpy(p, fname);
    fname = p;

    p = strrchr(fname + strlen(fname) - 4, '.');
    if(p) {
        strcpy(p, ".hed");
    } else {
        strcat(fname, ".hed");
    }

    if(!stricmp(cmd, "e")) {
        hedwad_extract(fname, folder);

    } else if(!stricmp(cmd, "l")) {
        list_only = 1;
        hedwad_extract(fname, folder);

    } else if(!stricmp(cmd, "b")) {
        hedwad_build(fname, folder);

    } else {
        printf("\n"
            "Error: wrong command\n"
            "use e for extracting the files in a wad archive into folder\n"
            "use b for building a wad file using the files contained in folder\n"
            "\n");
        exit(1);
    }

    printf("\n- %u files extracted\n\n", filez);
    return(0);
}



void hedwad_extract(u8 *fname, u8 *folder) {
    u64     offset;
    u32     length  = 0;
    u8      filename[FNAMEMAX + 1];

    printf("- read header file %s\n", fname);
    fd_hed = fopen(fname, "rb");
    if(!fd_hed) std_err();

    strcpy(strrchr(fname, '.'), ".wad");

    printf("- read wad file %s\n", fname);
    fd_wad = fopen(fname, "rb");
    if(!fd_wad) std_err();

    printf("- enter in folder %s\n", folder);
    if(chdir(folder) < 0) std_err();

    printf(OFFSZNAME);

    for(filez = 0; !feof(fd_hed); filez++) {
        offset = get32(fd_hed);
        if(offset == 0xffffffff) break;

        // try to guess because most of the archives have 2048 while others have 1
        if(filez == 1) {
            if(((u32)offset < (u32)(offset * SEEKSZ)) && ((u32)offset > length)) SEEKSZ = 1;
        }

        offset *= SEEKSZ;
        offset &= 0xffffffff;
        length = get32(fd_hed);
        getfn(fd_hed, filename, sizeof(filename) - 1);

        fname = create_dir(filename);
        //printf(" %08x%08x  %08x  %s\n", PRINTF64(offset), length, fname);
        printf(" %08x  %08x  %s\n", (u32)offset, length, fname);

        if(fseek(fd_wad, offset, SEEK_SET)) std_err();

        if(list_only) continue;
        copy_file(fd_wad, NULL, fname, length);
    }

    fclose(fd_hed);
    fclose(fd_wad);
}



void hedwad_build(u8 *fname, u8 *folder) {
    u8      filedir[FNAMEMAX + 1];

    printf("- write header file %s\n", fname);
    fd_hed = fopen(fname, "rb");
    if(fd_hed) {
        fclose(fd_hed);
        printf("\nError: the output file already exists, delete it and relaunch this tool\n\n");
        exit(1);
    }
    fd_hed = fopen(fname, "wb");
    if(!fd_hed) std_err();

    strcpy(strrchr(fname, '.'), ".wad");

    printf("- write wad file %s\n", fname);
    fd_wad = fopen(fname, "rb");
    if(fd_wad) {
        fclose(fd_wad);
        printf("\nError: the output file already exists, delete it and relaunch this tool\n\n");
        exit(1);
    }
    fd_wad = fopen(fname, "wb");
    if(!fd_wad) std_err();

    printf("- enter in folder %s\n", folder);
    if(chdir(folder) < 0) std_err();

    printf(OFFSZNAME);

    filez = 0;
    strcpy(filedir, ".");
    if(recursive_dir(filedir) < 0) {
        printf("\nError: an error occurred during the directory scanning\n");
        exit(1);
    }

    put32(fd_hed, 0xffffffff);

    fclose(fd_hed);
    fclose(fd_wad);
}



u32 get32(FILE *fd) {
    int     n1, n2, n3, n4;

    n1 = fgetc(fd);
    n2 = fgetc(fd);
    n3 = fgetc(fd);
    n4 = fgetc(fd);
    if((n1 < 0) || (n2 < 0) || (n3 < 0) || (n4 < 0)) std_err();
    return(n1 | (n2 << 8) | (n3 << 16) | (n4 << 24));
}



void put32(FILE *fd, u32 num) {
    fputc((num      ) & 0xff, fd);
    fputc((num >>  8) & 0xff, fd);
    fputc((num >> 16) & 0xff, fd);
    fputc((num >> 24) & 0xff, fd);
}



void getfn(FILE *fd, u8 *dst, u32 dstlen) {
    u32     i,
            n;

    for(i = 0; i < dstlen; ) {
        n = fgetc(fd);
        if((int)n < 0) break;
        dst[i] = n;
        i++;
        if(!n && !(i & 3)) break;
    }
}



void putfn(FILE *fd, u8 *name) {
    u32     i;

    while((*name == '.') || (*name == '/') || (*name == '\\')) name++;

    fputc('\\', fd);

    for(i = 1; ; i++) {
        if(*name == '/') {
            fputc('\\', fd);
        } else {
            fputc(*name, fd);
        }
        if(!*name) break;
        name++;
    }
    for(++i; i & 3; i++) fputc(0, fd);
}



void copy_file(FILE *fd, FILE *fdo, u8 *fname, u32 flen) {
    u32     rlen;
    u8      buff[BUFFSZ];

    if(!fdo) {
        fdo = fopen(fname, "wb");
        if(!fdo) std_err();

        for(rlen = sizeof(buff); flen; flen -= rlen) {
            if(flen < rlen) rlen = flen;
            fread(buff,  rlen, 1, fd);
            fwrite(buff, rlen, 1, fdo);
        }

        fclose(fdo);

    } else if(!fd) {
        fd = fopen(fname, "rb");
        if(!fd) std_err();

        while((rlen = fread(buff, 1, sizeof(buff), fd))) {
            fwrite(buff, rlen, 1, fdo);
        }

        fclose(fd);
    }
}



int build(u8 *fname, u32 length) {
    u64     offset;

    for(offset = ftell(fd_wad); offset % SEEKSZ; offset++) {
        fputc(0, fd_wad);
    }
    //printf(" %08x%08x  %08x  %s\n", PRINTF64(offset), length, fname);
    printf(" %08x  %08x  %s\n", (u32)offset, length, fname);

    put32(fd_hed, offset / SEEKSZ);
    put32(fd_hed, length);
    putfn(fd_hed, fname);

    copy_file(NULL, fd_wad, fname, 0);

    filez++;
    return(0);
}



u8 *create_dir(u8 *name) {
    u8      *p,
            *l;

    for(p = name; (*p == '\\') || (*p == '/') || (*p == '.'); p++);
    name = p;

    for(;;) {
        if((p[0] && (p[1] == ':')) || ((p[0] == '.') && (p[1] == '.'))) p[0] = '_';

        l = strchr(p, '\\');
        if(!l) {
            l = strchr(p, '/');
            if(!l) break;
        }
        *l = 0;
        p = l + 1;

        MKDIR(name);
        *l = PATHSLASH;
    }
    return(name);
}



int recursive_dir(u8 *filedir) {
    int     plen,
            ret     = -1;

#ifdef WIN32
    static int      winnt = -1;
    OSVERSIONINFO   osver;
    WIN32_FIND_DATA wfd;
    HANDLE  hFind;

    if(winnt < 0) {
        osver.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
        GetVersionEx(&osver);
        if(osver.dwPlatformId >= VER_PLATFORM_WIN32_NT) {
            winnt = 1;
        } else {
            winnt = 0;
        }
    }

    plen = strlen(filedir);
    strcpy(filedir + plen, "\\*.*");
    plen++;

    if(winnt) { // required to avoid problems with Vista and Windows7!
        hFind = FindFirstFileEx(filedir, FindExInfoStandard, &wfd, FindExSearchNameMatch, NULL, 0);
    } else {
        hFind = FindFirstFile(filedir, &wfd);
    }
    if(hFind == INVALID_HANDLE_VALUE) return(0);
    do {
        if(!strcmp(wfd.cFileName, ".") || !strcmp(wfd.cFileName, "..")) continue;

        strcpy(filedir + plen, wfd.cFileName);

        if(wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
            if(recursive_dir(filedir) < 0) goto quit;
        } else {
            build(filedir + 2, wfd.nFileSizeLow);
        }
    } while(FindNextFile(hFind, &wfd));
    ret = 0;

quit:
    FindClose(hFind);
#else
    struct  stat    xstat;
    struct  dirent  **namelist;
    int     n,
            i;

    n = scandir(filedir, &namelist, NULL, NULL);
    if(n < 0) {
        if(stat(filedir, &xstat) < 0) {
            printf("**** %s", filedir);
            std_err();
        }
        build(filedir + 2, xstat.st_size);
        return(0);
    }

    plen = strlen(filedir);
    strcpy(filedir + plen, "/");
    plen++;

    for(i = 0; i < n; i++) {
        if(!strcmp(namelist[i]->d_name, ".") || !strcmp(namelist[i]->d_name, "..")) continue;
        strcpy(filedir + plen, namelist[i]->d_name);

        if(stat(filedir, &xstat) < 0) {
            printf("**** %s", filedir);
            std_err();
        }
        if(S_ISDIR(xstat.st_mode)) {
            if(recursive_dir(filedir) < 0) goto quit;
        } else {
            build(filedir + 2, xstat.st_size);
        }
        free(namelist[i]);
    }
    ret = 0;

quit:
    for(; i < n; i++) free(namelist[i]);
    free(namelist);
#endif
    filedir[plen - 1] = 0;
    return(ret);
}



void std_err(void) {
    perror("\nError");
    exit(1);
}


